declare type MapToTyped<T> = { [name: string]: T }

declare type MapToAny = { [name: string]: any }

declare type MapToSimple = MapToTyped<string | number | boolean>

declare interface TraceContext {

}

declare interface JavaMap {

}

declare interface Date {
  getTime: () => number
}

declare interface ScriptBase {
  traceContext: () => TraceContext
  logDebug: (tc: TraceContext, msg: string) => void
  logInfo: (tc: TraceContext, msg: string) => void
  logError: (tc: TraceContext, msg: string, err: any) => void
  scheduledAtFixedDelay: (name: string, delay: number, work: (counter: number) => void) => void
  robustExecute: (tc: TraceContext, description: string, func: string, args: MapToSimple) => void
  runOnce: (tc: TraceContext, mainId: string, actionId: string, work: () => MapToAny) => MapToAny
  withLock: (name: string, action: () => void) => void
  parseJsonString: (str: string) => MapToAny
  jsonToString: (o: any) => string | null
  xmlToJsonStr: (o: string | number) => any | null
  now: () => Date
  throwBzError: (code: string, args: any[]) => any
  assignJavaMapToJsMap: (jsMap: MapToAny, javaMap: JavaMap) => void
  setScriptButtons: (buttons: string) => void
  getGlobalValue: (key: String) => any | null
  setGlobalValue: (key: String, value: any | null) => void
}

declare const base: ScriptBase

declare interface ScriptThread {
  createThread: (name: string, func: string) => void
  sleep: (ms: number) => void
  interrupted: () => boolean
}

declare const thread: ScriptThread

declare interface HttpServer {
  registerHandler: (method: string, path: string, func: string, auth: boolean) => void
}

declare interface HttpServerContext {
  getBodyAsString: () => string
  setJson: (jo: any) => void
}

declare const httpServer: HttpServer

declare interface HttpClient {
  requestJson: (method: string, url: string, reqBody: any | null, headers: MapToTyped<string> | null) => HttpResult
  /**
   * 反复轮训直到成功。使用此方法的好处：不用自己写循环。优化日志，开始结束时打印日志。中间如果一致是相同结果，不重复打印日志。
   * purpose 是目的，写好了方便差错。
   * 后面会与 SOC 结合。
   */
  requestJsonUntil: (method: string, url: string, reqBody: any | null, headers: MapToTyped<string> | null,
                     purpose: string, delay: number, check: (r: HttpResult) => boolean) => HttpResult
  buildRetryCfg: (async: boolean, maxRetryNum: number, retryDelay: number, extFunction: string) => ExtCallRetryCfg
  requestJsonWithRetry: (method: string, url: string, reqBody?: any, headers?: MapToTyped<string>, retryCfg?: ExtCallRetryCfg | null) => HttpResult
  requestWebService: (method: string, url: string, reqBody?: any, headers?: MapToTyped<string>, auth?: MapToTyped<string>) => HttpResult
}

declare interface ExtCallRetryCfg {
  async: boolean
  maxRetryNum: number
  retryDelay: number
  extFunction: string
}

declare const httpClient: HttpClient

declare interface HttpResult {
  successful: boolean
  code: number
  bodyString?: string | null
}

declare type SocAttention = "None" | "Green" | "Yellow" | "Red"

declare interface Soc {
  updateStringNode: (id: string, desc: string, content: string, attention: SocAttention) => void
  updateIntNode: (id: string, desc: string, content: number, attention: SocAttention) => void
  updateJsonNode: (id: string, desc: string, content: any, attention: SocAttention) => void
  getNode: (id: string) => SocNode | null
  removeNode: (id: string) => void
}

declare const soc: Soc

declare interface SocNode {
  getId(): string

  getLabel(): string

  getValue(): any

  getAttention(): SocAttention
}

declare interface ComplexQuery {
}

declare interface Cq {
  all: () => ComplexQuery
  and: (items: ComplexQuery[]) => ComplexQuery
  or: (items: ComplexQuery[]) => ComplexQuery
  eq: (field1: string, v: any) => ComplexQuery
  ne: (field1: string, v: any) => ComplexQuery
  lt: (field1: string, v: any) => ComplexQuery
  lte: (field1: string, v: any) => ComplexQuery
  gt: (field1: string, v: any) => ComplexQuery
  gte: (field1: string, v: any) => ComplexQuery
  idEq: (id: string) => ComplexQuery
  include: (field1: string, items: any[]) => ComplexQuery
  empty: (field1: string) => ComplexQuery
  notEmpty: (field1: string) => ComplexQuery
}

declare const cq: Cq

declare interface ScriptEntity {
  createOne: (tc: TraceContext, entityName: string, evJson: MapToAny, o: CreateOptions | null) => string
  createMany: (tc: TraceContext, entityName: string, evJsonList: MapToAny[], o: CreateOptions | null) => string[]
  buildCreateOptions: (keepId: boolean) => CreateOptions
  updateOne: (tc: TraceContext, entityName: string, queryJson: ComplexQuery, updateJson: MapToAny) => number
  updateOneById: (tc: TraceContext, entityName: string, id: string, updateJson: MapToAny) => number
  updateMany: (tc: TraceContext, entityName: string, queryJson: ComplexQuery, updateJson: MapToAny, o: UpdateOptions | null) => number
  buildUpdateOptions: (limit: number) => UpdateOptions
  removeOne: (tc: TraceContext, entityName: string, queryJson: ComplexQuery) => number
  removeMany: (tc: TraceContext, entityName: string, queryJson: ComplexQuery, o: RemoveOptions | null) => number
  buildRemoveOptions: (limit: number) => RemoveOptions
  count: (tc: TraceContext, entityName: string, queryJson: ComplexQuery) => number
  findOne: (tc: TraceContext, entityName: string, queryJson: ComplexQuery, o: FindOptions | null) => MapToAny | null
  findOneById: (tc: TraceContext, entityName: string, id: string, o: FindOptions | null) => MapToAny | null
  findMany: (tc: TraceContext, entityName: string, queryJson: ComplexQuery, o: FindOptions | null) => MapToAny[]
  exists: (tc: TraceContext, entityName: string, queryJson: ComplexQuery) => boolean
  buildFindOptions: (projection: string[] | null, sort: string[] | null, skip: number | null, limit: number | null) => FindOptions
  clearCacheAll: () => void
  clearCacheByEntity: (entityName: string) => void
}

declare const entity: ScriptEntity

declare interface CreateOptions {
  keepId?: boolean
}

declare interface UpdateOptions {
  limit?: number
}

declare interface RemoveOptions {
  limit?: number
}

declare interface FindOptions {
  projection?: string[]
  sort?: string[]
  skip?: number
  limit?: number
}

declare interface ScriptEntityExt {
  /**
   * 注意实体生命周期拦截器：实体新建前
   */
  extBeforeCreating: (entityName: string, func: string) => void

  /**
   * 注意实体生命周期拦截器：实体新建后
   */
  extAfterCreating: (entityName: string, func: string) => void

  /**
   * 注意实体生命周期拦截器：实体更新前
   */
  extBeforeUpdating: (entityName: string, func: string) => void

  /**
   * 注意实体生命周期拦截器：实体更新后
   */
  extAfterUpdating: (entityName: string, func: string) => void

  /**
   * 实体删除前
   */
  extBeforeRemoving: (entityName: string, func: string) => void

  /**
   * 实体删除后
   */
  extAfterRemoving: (entityName: string, func: string) => void
}

declare const entityExt: ScriptEntityExt

declare interface ScriptFalcon {
  runTaskByLabelAsync: (tc: TraceContext, defLabel: string, jsInputParams: MapToAny) => string
}

declare const falcon: ScriptFalcon


declare interface ScriptWcs {
  /**
   * 机器人是否在线
   */
  isRobotOnline: (robotName: string) => boolean
  mustGetRobotInfoAll: (id: string) => MrRobotInfoAll
  /**
   * 寻找离机器人最近的通讯点
   */
  findClosestRobotConnectedPoint: (robotName: string) => MapToAny | null
  buildHaiMoves: (tc: TraceContext, robotId: string, rawSteps: string) => MapToAny[]
  buildSeerMoves: (tc: TraceContext, robotId: string, rawSteps: string) => MapToAny[]
  /**
   * 未禁用、可接单、在线的机器人
   */
  listWorkableRobots: () => MrRobotInfoAll[]
  /**
   * 未禁用、可接单、在线、空闲（扩展任务状态）的机器人
   */
  listExtIdleRobots: () => MrRobotInfoAll[]
  /**
   * 采用 RunOnce 的方式，创建直接运单，并等待其完成
   */
  awaitRunOnceDirectRobotOrder: (tc: TraceContext, robotName: string, mainId: string, action: string, moves: MapToAny[]) => string
  mustGetBinRobotArgs: (bin: string, action: string) => BinRobotArgs
  mustGet3051Params: (bin: string, action: string) => MapToAny
  unlockByRobot: (tc: TraceContext, robotId: string) => void
  tryLockOneSiteByName: (tc: TraceContext, robotId: string, siteIds: string[]) => string | null
  /**
   * 重置所有扩展机器人，终止所有后台任务，清楚所有地图资源锁
   */
  resetExtRobots: (tc: TraceContext) => void
}

declare const wcs: ScriptWcs

declare interface ScriptPlc {
  modbusRead: (tc: TraceContext, deviceName: string, reqMap: ModbusReadReq) => number[]
  /**
   * 重复读，直到等于某个值；无限尝试
   */
  modbusReadUtilEq: (tc: TraceContext, deviceName: String, reqMap: ModbusReadReq, targetValue: number, readDelay: number) => void
  /**
   * 重复读，直到等于某个值；最多尝试指定次数
   */
  modbusReadUtilEqMaxRetry: (tc: TraceContext, deviceName: String, reqMap: ModbusReadReq, targetValue: number, readDelay: number, maxRetry: number) => boolean
  modbusWrite: (tc: TraceContext, deviceName: string, reqMap: ModbusWriteReq, values: number[]) => void

  s7Read: (tc: TraceContext, deviceName: string, reqMap: S7ReadReq) => any
  s7ReadUntilEq: (tc: TraceContext, deviceName: string, reqMap: S7ReadReq, targetValue: any, readDelay: number) => void
  s7Write: (tc: TraceContext, deviceName: string, reqMap: S7WriteReq) => void
}

declare const plc: ScriptPlc

declare interface ModbusReadReq {
  code: number
  address: number
  qty: number
  slaveId?: number
  maxRetry?: number
  retryDelay?: number
}

declare interface ModbusWriteReq {
  code: number
  address: number
  slaveId?: number
  maxRetry?: number
  retryDelay?: number
}

declare interface S7ReadReq {
  blockType: "DB" | "Q" | "I" | "M" | "V"
  dataType: "BOOL" | "BYTE" | "INT16" | "UINT16" | "INT32" | "UINT32" | "FLOAT32" | "FLOAT64" | "STRING"
  dbId: number
  byteOffset: number
  bitOffset?: number
  maxRetry?: number
  retryDelay?: number
}

declare interface S7WriteReq {
  value: any
  blockType: "DB" | "Q" | "I" | "M" | "V"
  dataType: "BOOL" | "BYTE" | "INT16" | "UINT16" | "INT32" | "UINT32" | "FLOAT32" | "FLOAT64" | "STRING"
  dbId: number
  byteOffset: number
  bitOffset?: number
  maxRetry?: number
  retryDelay?: number
}

declare interface BinRobotArgs {
  bin: string
  action: string
  site: string
  fields: MapToAny
  rbkFields: MapToAny
}

declare interface MrRobotInfoAll {
  id: string
  systemConfig: MapToAny
  runtimeRecord: MapToAny | null
  selfReport: MrRobotSelfReport | null
}

declare interface MrRobotSelfReport {
  error?: boolean
  errorMsg?: string | null
  main?: MrRobotSelfReportMain | null
  rawReport?: MapToAny | null
  timestamp?: any
}

declare interface MrRobotSelfReportMain {
  battery: number | null
  x: number | null
  y: number | null
  direction: number | null
  currentSite: string | null
  blocked: boolean | null
  charging: boolean | null
  currentMap: string | null
  currentMapMd5: string | null
  alerts?: MrRobotAlert[]
}

declare interface MrRobotAlert {
  level: string
  code: string | null
  message: string
  times: number | null
  timestamp: any
}

declare interface ScriptUi {
  /**
   * 在指定工位电脑上，打开业务对象的界面
   */
  openEntityViewPage: (workSite: string, entityName: string, id: string, mode: "Edit" | "Read") => void

}

declare const ui: ScriptUi

// ======== BZ

declare interface Bz {
  tryCallContainer: (tc: TraceContext) => void
  tryOutbound: (tc: TraceContext) => void
  finishPutOrderByContainer: (tc: TraceContext, containerId: string) => void
  /**
   * 修复并创建库存明细。填充物料信息、位置信息、容器等。
   */
  fixCreateInvLayout: (tc: TraceContext, layouts: MapToAny[]) => void
}

declare const bz: Bz

declare interface OrderLineSource {
  orderId: string
  lineId: string
  lineNo: number
}

declare type ShortOption = "Stop" | "Part"

declare interface CallContainerPlan {
  mixKey: string
  materialId: string
  feature: JavaMap, // 额外的筛选字段
  restQty: number // 待装数量
  assignedQty: number // 实际成功分配的数量
  qty: number, // 入库数
  callContainerQty: number, // 已叫
  sources: OrderLineSource[] // 来源单据单行
}

declare interface MaterialContainerMaxQty {
  containerType: string
  subNum: number
  maxQty: number
}

declare interface ResLock {
  withLock: (action: () => void) => void
  // isLocked: (tc: TraceContext, resType: string, resId: string) => boolean
  // getOwner: (tc: TraceContext, resType: string, resId: string) => string | null
  // tryLockRes: (tc: TraceContext, resType: string, resId: string, owner: string, reason: string) => boolean
  // unlockRes: (tc: TraceContext, resType: string, resId: string) => void
  // unlockResIfLockedBy: (tc: TraceContext, resType: string, resId: string, me: string) => Boolean
  // listMyRes: (tc: TraceContext, me: string) => MapToAny[]
}

declare const resLock: ResLock

declare interface EntityMeta {
  getName: () => string
}

declare interface CallEntityExtButtonReq {
  func: string // 调用函数的名字
  entityName: string // 业务对象名
  selectedIds?: string[] | null // 选中的业务对象 ID，用于列表界面
  evId?: string | null // 当前业务对象 ID，用于编辑和查看界面
  ev?: MapToAny | null // 当前业务对象值，用于新增编辑和查看界面
}


declare interface ScriptUtils {
  isNullOrBlank: (str: string | null | undefined) => Boolean
  substringAfter: (str: string, sep: string) => string
  substringBefore: (str: string, sep: string) => string
  splitTrim: (str: string | null | undefined, sep: string) => string[]
  //
  anyToInt: (v: any) => number | null
  anyToLong: (v: any) => number | null
  anyToFloat: (v: any) => number | null
  anyToDouble: (v: any) => number | null
  //
  anyToBool: (a: any) => boolean
  //
  anyToDate: (input: any) => Date
  /**
   * format 是 Java 的格式
   */
  formatDate: (d: Date, format: string) => string
  //
  /**
   * 产生 UUID 字符串
   */
  uuidStr: () => string
  /**
   * 产生 ObjectId 字符串
   */
  oidStr: () => string
}

declare const utils: ScriptUtils

// =======
// 真实代码

interface ScriptButton {
  label: string
  func: string
  confirmText?: string | null
  callTimeout?: number | null
  inputEntityName?: string | null
  inputMaxWidth?: number | null
}

function setScriptButtons(buttons: ScriptButton[]) {
  const buttonsStr = JSON.stringify(buttons)
  base.setScriptButtons(buttonsStr)
}

interface ScriptButtonResult {
  message?: string | null
}